--[[	QuickFrames - A lightweight fork on QuickFrames by Dolmur by Consensual Lovin @ Entity

Purpose: 	To provide lightweight raidframes designed with DPS in mind to show status of the
			raid and interrupts. Forked heavily from DolmurFrames/QuickFrames who has since
			stopped maintaining his project. The code will be heavily commented to encourage
			further forks but will maintain the aesthetic style of QuickFrames.
]]
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- QuickFrames Module Definition
-----------------------------------------------------------------------------------------------
local QuickFrames = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local kDebug = false

local kVersion = 1.5

local knDefaultFrameWidth = 195
local knDefaultNumColumns = 1
local knFramePadding = 5
local knFrameHeight = 25 -- must correspond to XML

local kColorDead = CColor.new(0,0,0,1)
local kColorAlive = CColor.new(0,0.5,0,1)
local kColorOffline = CColor.new(0,0,0,1)
local kTextColorDead = CColor.new(1,0,0,1)
local kTextColorAlive = CColor.new(1,1,1,1)
local kTextColorOffline = CColor.new(1,1,1,1)

local kDefaultOpacity = 0.6
local kHighlightedOpacity = 0.9
local knUpdateTimer = 0.5

local kReadySprite = "CRB_Raid:sprRaid_Icon_ReadyCheckDull"
local kNotReadySprite = "CRB_Raid:sprRaid_Icon_NotReadyDull"

local kNoBuffsSprite = "CRB_ChatLogSprites:btnCL_EmotesDisabled"
local kPartialBuffsSprite = "CRB_ChatLogSprites:btnCL_EmotesNormal"
local kFullBuffsSprite = "CRB_ChatLogSprites:btnCL_EmotesFlyby"

local ktCommonFirstNameTitles = {
	"Engineer", 
	"Doctor", 
	"Medic", 
	"Spellslinger", 
	"Warrior", 
	"Esper", 
	"Stalker", 
	"Captain", 
	"Hung", 
	"Brother", 
	"Mr", 
	"Not",
	}

local ktIdToClassColors = {
	[GameLib.CodeEnumClass.Esper]			= CColor.new(0.1, 0.5, 0.7),
	[GameLib.CodeEnumClass.Medic] 			= CColor.new(0.2, 0.6, 0.1),
	[GameLib.CodeEnumClass.Stalker] 		= CColor.new(0.5, 0.1, 0.8),
	[GameLib.CodeEnumClass.Warrior] 		= CColor.new(0.8, 0.1, 0.1),
	[GameLib.CodeEnumClass.Engineer] 		= CColor.new(0.65, 0.65, 0),
	[GameLib.CodeEnumClass.Spellslinger] 	= CColor.new(0.9, 0.4, 0)
}

local kstrRaidMarkerToSprite = {
	"Icon_Windows_UI_CRB_Marker_Bomb",
	"Icon_Windows_UI_CRB_Marker_Ghost",
	"Icon_Windows_UI_CRB_Marker_Mask",
	"Icon_Windows_UI_CRB_Marker_Octopus",
	"Icon_Windows_UI_CRB_Marker_Pig",
	"Icon_Windows_UI_CRB_Marker_Chicken",
	"Icon_Windows_UI_CRB_Marker_Toaster",
	"Icon_Windows_UI_CRB_Marker_UFO",
}

local ePlayerStatus_Alive = 0
local ePlayerStatus_Dead = 1
local ePlayerStatus_Offline = 2

local knReadyCheckTimeout = 60
local knHideReadyCheckTimer = 10

local knXCursorOffset = 10
local knYCursorOffset = 25

local eQuickActionsMenu_ReadyCheck = 0
local eQuickActionsMenu_Lock = 1
local eQuickActionsMenu_LootOptions = 2
local eQuickActionsMenu_LeaderOptions = 3
local eQuickActionsMenu_MassRaidAssist = 4
local eQuickActionsMenu_SetRole = 5
local eQuickActionsMenu_RoleDPS = 6
local eQuickActionsMenu_RoleHealer = 7
local eQuickActionsMenu_RoleTank = 8

local kRoleActionsMap = {
	[eQuickActionsMenu_RoleDPS] = { strLabel = "DPS" },
	[eQuickActionsMenu_RoleHealer] = { strLabel = "Healer" },
	[eQuickActionsMenu_RoleTank] = { strLabel = "Tank" },
}

local kQuickActionsMap = {
	[eQuickActionsMenu_ReadyCheck] = { strLabel = "Ready Check" },
	[eQuickActionsMenu_Lock] = { strLabel = "Toggle Lock" },
	[eQuickActionsMenu_LootOptions] = { strLabel = "Loot Options", bLeaderOnly = true },
	[eQuickActionsMenu_LeaderOptions] = { strLabel = "Leader Options", bLeaderOnly = true },
	[eQuickActionsMenu_MassRaidAssist] = { strLabel = "Raid Assist All", bLeaderOnly = true },
	[eQuickActionsMenu_SetRole ] = { strLabel = "Set Role", tSubMenuMap = kRoleActionsMap },
}

-- Settings
local eOptionType_CheckBox				= 0
local eOptionType_Slider				= 1

local eOption_Version					= 0
local eOption_DefaultOpacity			= 1
local eOption_HighlightedOpacity		= 2
local eOption_ShowMarks					= 3
local eOption_ShowBuffStatus			= 4
local eOption_ShowQuickActions			= 5
local eOption_Locked					= 6
local eOption_ShowVitals				= 7
local eOption_ClassColors				= 8
local eOption_FrameWidth				= 9
local eOption_NumColumns				= 10
local eOption_UpdateTimer				= 11

local kSettingsDataMap = {
	["DefaultOpacity"]				= { eOption = eOption_DefaultOpacity,			eType = eOptionType_Slider },
	["HighlightedOpacity"]			= { eOption = eOption_HighlightedOpacity,		eType = eOptionType_Slider },
	["UpdateTimer"]					= { eOption = eOption_UpdateTimer,				eType = eOptionType_Slider },
	["FrameWidth"]					= { eOption = eOption_FrameWidth,				eType = eOptionType_Slider,		bInteger = true },
	["NumColumns"]					= { eOption = eOption_NumColumns,				eType = eOptionType_Slider,		bInteger = true },
	
	["Marks"]						= { eOption = eOption_ShowMarks,				eType = eOptionType_CheckBox },
	["BuffStatus"]					= { eOption = eOption_ShowBuffStatus,			eType = eOptionType_CheckBox },
	["QuickActions"]				= { eOption = eOption_ShowQuickActions,			eType = eOptionType_CheckBox },
	["Vitals"]						= { eOption = eOption_ShowVitals,				eType = eOptionType_CheckBox },
	["ClassColors"]					= { eOption = eOption_ClassColors,				eType = eOptionType_CheckBox },
}

local kFoodBuffStr = "Stuffed!"
local kBoostStrList = {
	"Brutality Boost",
	"Finesse Boost",
	"Moxie Boost",
	"Insight Boost",
	"Tech Boost",
	"Grit Boost",
	"Reactive Moxie Boost",
	"Ractive Finesse Boost",
	"Reactive Insight Boost",
	"Reactive Avoidance Boost",
	"Reactive Armor Boost",
	"Reactive Brutality Boost",
	"Reactive Tech Boost",
	"Reactive Protection Boost"
}

local ktRoleDictionary = { "Tanks", "Healers", "DPS"}

local knFoodBuffOffset = -25
local knBoostOffset = -47
local knReadyCheckWidth = 20

-----------------------------------------------------------------------------------------------
-- QuickFrames Addon Initialization
-----------------------------------------------------------------------------------------------
function QuickFrames:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    return o
end

function QuickFrames:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function QuickFrames:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("QuickFrames.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	Apollo.LoadSprites("QuickFramesSprites.xml", "QuickFramesSprites")
end

function QuickFrames:OnDocLoaded()
	if not self.xmlDoc or not self.xmlDoc:IsLoaded() then
		return
	end
	
	Apollo.RegisterSlashCommand("qf", "OnChatInvoke", self)
	
	Apollo.RegisterEventHandler("Group_Join", 								"OnGroupJoin", self)
	Apollo.RegisterEventHandler("Group_Left", 								"OnGroupLeft", self)
	Apollo.RegisterEventHandler("Group_Add",								"OnGroupAdd", self)
	Apollo.RegisterEventHandler("Group_Remove",								"OnGroupRemove", self)
	Apollo.RegisterEventHandler("Group_FlagsChanged",						"OnGroupFlagsChanged", self)	
	Apollo.RegisterEventHandler("Group_ReadyCheck",							"OnReadyCheck", self)
	Apollo.RegisterEventHandler("Group_SetMark",							"OnSetMark", self)
	Apollo.RegisterEventHandler("Group_MemberFlagsChanged",					"OnMemberFlagsChanged", self)
	Apollo.RegisterEventHandler("Group_MemberPromoted",						"OnMemberPromoted", self)

	self.wndRaidFrame = Apollo.LoadForm(self.xmlDoc, "RaidFrame", nil, self)
	self.wndRaidFrame:Show(false, true)
	self.wndSettings = Apollo.LoadForm(self.xmlDoc, "SettingsForm", nil, self)
	self.wndSettings:Show(false, true)
	self:InitializeSettings()

	-- Register Timers
	self.readyCheckTimer = ApolloTimer.Create(knReadyCheckTimeout, false, "OnReadyCheckTimeout", self)
	self.hideReadyCheckTimer = ApolloTimer.Create(knHideReadyCheckTimer, false, "HideReadyCheckIcons", self)
	self.updateTimer = ApolloTimer.Create(self.tSettings[eOption_UpdateTimer], true, "OnTimer", self)

	-- Stop Timers
	self.readyCheckTimer:Stop()
	self.hideReadyCheckTimer:Stop()
	self.updateTimer:Stop()

	self.wndQuickActionsBtn = Apollo.LoadForm(self.xmlDoc, "QuickActions", nil, self)
	self.wndQuickActionsBtn:Show(false, true)
	
	self.tGroups = { }
	self.nGroupCount = 0
	self.wndDefaultAddGroup = nil
	self.tRaidMembers = { }
	self.tSortedMembers = { }
	self.tCurrentlyMarked = { }
	self.nNumReadyCheckResponses = -1
	self.bShowInterrupts = false
	self.bHovering = false
	
	if GroupLib.InRaid() then
		self:InitializeRaidGroup()
	end
end

function QuickFrames:InitializeSettings()
	local tDefaultSettings = {
		[eOption_Version]		= kVersion,
		[eOption_DefaultOpacity] = kDefaultOpacity,
		[eOption_HighlightedOpacity] = kHighlightedOpacity,
		[eOption_FrameWidth] = knDefaultFrameWidth,
		[eOption_NumColumns] = knDefaultNumColumns,
		[eOption_UpdateTimer] = knUpdateTimer,
		
		[eOption_ShowMarks] = true,
		[eOption_ShowBuffStatus] = true,
		[eOption_ShowQuickActions] = true,
		[eOption_ClassColors] = true,
		
		[eOption_Locked] = true
	}

	if self.tRenamedCharacters == nil then
		self.tRenamedCharacters = { }
	end

	if self.tSettings == nil or self.tSettings[eOption_Version] ~= kVersion then
		self.tSettings = tDefaultSettings
	else
		for k, v in pairs(tDefaultSettings) do
			if self.tSettings[k] == nil then
				self:DebugPrint("Added Missing Setting: " .. k)
				self.tSettings[k] = v
			end
		end
	end
	
	if self.tRaidWindowOffsets ~= nil then
		self.wndRaidFrame:SetAnchorOffsets(unpack(self.tRaidWindowOffsets))
	end
		
	local bLeader = GroupLib.AmILeader()
	local arOptions = self.wndSettings:FindChild("OptionsFrame"):GetChildren()
	for i = 1, #arOptions do
		local strName = arOptions[i]:GetName()
		local tOptionData = kSettingsDataMap[strName]
		if tOptionData then
			arOptions[i]:SetData(tOptionData.eOption)
			if tOptionData.eType == eOptionType_CheckBox then
				arOptions[i]:SetCheck(self.tSettings[tOptionData.eOption])
			elseif tOptionData.eType == eOptionType_Slider then
				local nDecimalPlaces = kSettingsDataMap[arOptions[i]:GetName()].bInteger and 0 or 2
				arOptions[i]:FindChild("SliderBar"):SetValue(self.tSettings[tOptionData.eOption])
				arOptions[i]:FindChild("EditBox"):SetText(string.format("%." .. nDecimalPlaces .. "f", self.tSettings[tOptionData.eOption]))
			end
		elseif strName == "RaidLeaderOptions" or strName == "LootOptions" then
			arOptions[i]:Enable(bLeader)
		end
	end
end

function QuickFrames:Reset()
	self.tSettings = { }
	self.tRaidWindowOffsets = {0, 0, 0, 0}
	self.tRenamedCharacters = { }
	self:InitializeSettings()
	self:DestroyAll()
	self:InitializeRaidGroup()
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Data Persistence
-----------------------------------------------------------------------------------------------
function QuickFrames:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	local tSave = {
		tRaidWindowOffsets = {self.wndRaidFrame:GetAnchorOffsets()},
		tSettings = self.tSettings,
		tRenamedCharacters = self.tRenamedCharacters
	}

	return tSave
end

function QuickFrames:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character or not tSavedData then
		return
	end
	if tSavedData ~= nil then
		self.tSettings = tSavedData.tSettings
		self.tRaidWindowOffsets = tSavedData.tRaidWindowOffsets
		self.tRenamedCharacters = tSavedData.tRenamedCharacters
	end
end
-----------------------------------------------------------------------------------------------
-- QuickFrames Main Functions
-----------------------------------------------------------------------------------------------
function QuickFrames:BuildDefaultFrames()
	for i = 1, #ktRoleDictionary do
		self.tGroups[i] = self:CreateGroup(ktRoleDictionary[i])
	end
	local bLeader = GroupLib.AmILeader()
	for i = 1, GroupLib.GetMemberCount() do
		local newMember = GroupLib.GetGroupMember(i)
		local tMemberData = self:CreateMemberData(newMember)
		self.tRaidMembers[newMember.strCharacterName] = tMemberData
	end
	self:UpdateMemberSort()
	self.wndRaidFrame:Show(true, true)
end

function QuickFrames:UpdateMemberSort()
	table.sort(self.tSortedMembers, function(a,b) return a[2] < b[2] end)
	for key, name in pairs(self.tSortedMembers) do
		self:AddMemberToGroup(self.tGroups[self.tRaidMembers[name[1]].nRole], self.tRaidMembers[name[1]], true)
	end
	self:FormatGroups()
end

function QuickFrames:BestName(strCharacterName)
	if self.tRenamedCharacters[strCharacterName] then
		return self.tRenamedCharacters[strCharacterName]
	end

	if string.len(strCharacterName) > 8 then
		for i in string.gmatch(strCharacterName, "%S+") do
			if string.len(i) > 3 then
				for key, title in pairs(ktCommonFirstNameTitles) do
					if i ~= title then
						return i
					end
				end
			end
		end
	end
	return strCharacterName
end

function QuickFrames:RenameCharacter(strCharacterName, newCharacterName)
 	for key, tMemberData in pairs(self.tRaidMembers) do
 		if string.lower(tMemberData.wndNameText:GetText()) == string.lower(strCharacterName) or
 		   string.lower(tMemberData.strCharacterName) == string.lower(strCharacterName) then
 			self.tRenamedCharacters[tMemberData.strCharacterName] = newCharacterName
 			tMemberData.wndNameText:SetText(newCharacterName)
 			self:InfoPrint("Found " .. strCharacterName .. " as " .. tMemberData.strCharacterName .. "!")
 			return
 		end
 	end
	self:InfoPrint("Character not found, please try again. [Usage: /rf rn [rename] <First Name> <Last Name> <New Name>")
end

function QuickFrames:AddMemberToGroup(wndGroup, tMemberData, bSkipFormatting)
	if not wndGroup then
		return
	end

	self:RemoveMemberFromGroup(tMemberData)
	tMemberData.wndGroup = wndGroup
	local groupMemberList = wndGroup:FindChild("MemberList")
	local wnd = Apollo.LoadForm(self.xmlDoc, "MemberItem", groupMemberList, self)
	
	tMemberData.wndMemberItem = wnd:FindChild("Contents")
	tMemberData.wndMemberItem:SetData(tMemberData.strCharacterName)
	tMemberData.wndBorder = tMemberData.wndMemberItem:FindChild("Border")
	tMemberData.wndBGBar = tMemberData.wndMemberItem:FindChild("BG")
	tMemberData.wndHealthBar = tMemberData.wndMemberItem:FindChild("HealthBar")
	tMemberData.wndShieldBar = tMemberData.wndMemberItem:FindChild("ShieldBar")
	tMemberData.wndAbsorbBar = tMemberData.wndMemberItem:FindChild("AbsorbBar")
	tMemberData.wndCurrInts = tMemberData.wndMemberItem:FindChild("IntText")
	tMemberData.wndNameText = wnd:FindChild("Text")

	if tMemberData.bIsLeader then 
		tMemberData.wndBorder:SetSprite("CRB_ActionBarIconSprites:sprActionBar_OrangeBorder")
		tMemberData.wndBorder:SetBGColor(CColor.new(1,1,0))
	end
	if tMemberData.bIsRaidAssist and not tMemberData.bIsLeader then
		tMemberData.wndBorder:SetSprite("CRB_ActionBarIconSprites:sprActionBar_OrangeBorder") 
		tMemberData.wndBorder:SetBGColor(CColor.new(0,.8,1))
	end

	if self.bShowInterrupts == false then tMemberData.wndCurrInts:Show(false, true) end
	
	-- Background color based on player status, opacity based on frame state
	if tMemberData.eStatus == ePlayerStatus_Offline then
		tMemberData.wndBGBar:SetBGColor(kColorOffline)
		tMemberData.wndNameText:SetTextColor(kTextColorOffline)
	elseif tMemberData.eStatus == ePlayerStatus_Alive then
		if self.tSettings[eOption_ClassColors] then
			tMemberData.wndBGBar:SetBGColor(ktIdToClassColors[tMemberData.eClassId])
		else
			tMemberData.wndBGBar:SetBGColor(kColorAlive)
			tMemberData.wndNameText:SetTextColor(kTextColorAlive)
		end
	else
		tMemberData.wndBGBar:SetBGColor(kColorDead)
		tMemberData.wndNameText:SetTextColor(kTextColorDead)
	end
	tMemberData.wndBGBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	tMemberData.wndHealthBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	tMemberData.wndShieldBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	tMemberData.wndAbsorbBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	tMemberData.wndNameText:SetText(self:BestName(tMemberData.strCharacterName))
	
	if tMemberData.nMarkerId then
		local wndMark = wnd:FindChild("Contents"):FindChild("Mark"):FindChild("Icon")
		wndMark:SetSprite(kstrRaidMarkerToSprite[tMemberData.nMarkerId])
		wndMark:Show(true, true)
		self.tCurrentlyMarked[tMemberData.nMarkerId] = tMemberData.strCharacterName
	end
	
	local bReadyResponse = false
	if tMemberData.wndMemberItem and tMemberData.wndMemberItem:FindChild("ReadyIcon"):IsShown() then
		bReadyResponse = true
	end
	
	self:DebugPrint("Adding: " .. tMemberData.strCharacterName .. " to " .. wndGroup:FindChild("GroupName"):GetData())
	
	if bReadyResponse then
		local wndReady = wnd:FindChild("Contents"):FindChild("ReadyIcon")
		wndReady:SetSprite(tMemberData.bReady and kReadySprite or kNotReadySprite)
		wndReady:Show(true, true)
	end
	
	if not bSkipFormatting then
		self:FormatGroups()
	end
end

function QuickFrames:RemoveMemberFromGroup(tMemberData)
	if tMemberData.wndMemberItem then
		tMemberData.wndMemberItem:GetParent():Destroy()
		tMemberData.wndMemberItem = nil
		self:FormatGroups()
	end
end

function QuickFrames:FormatGroups()
	local numTotalMembers = 0
	for i = 1, #self.tGroups do
		numTotalMembers = numTotalMembers + #self.tGroups[i]:FindChild("MemberList"):GetChildren()
		self:FormatGroup(self.tGroups[i], true)
	end
	local nTotalWidth = self.tSettings[eOption_FrameWidth] * self.tSettings[eOption_NumColumns] + 15
	local nTotalHeight = (knFrameHeight + knFramePadding) * math.max(math.ceil(numTotalMembers / self.tSettings[eOption_NumColumns]), 3)
	local nLeft, nTop, nRight, nBottom = self.wndRaidFrame:GetAnchorOffsets()
	self.wndRaidFrame:SetAnchorOffsets(nLeft, nTop, nLeft + nTotalWidth, nTop + nTotalHeight + 15 * #self.tGroups)
	self.wndRaidFrame:ArrangeChildrenVert(0)
end

function QuickFrames:FormatGroup(wndGroup, bWidthCalc)
	local numGroupMembers = #wndGroup:FindChild("MemberList"):GetChildren()
	if numGroupMembers == 0 then
		wndGroup:Show(false, true)
	else
		wndGroup:Show(true, true)
	end
	local nLeft, nTop, nRight, nBottom = wndGroup:GetAnchorOffsets()
	local nGroupWidth = self.tSettings[eOption_FrameWidth] * self.tSettings[eOption_NumColumns]
	local nGroupHeight = knFrameHeight * math.ceil(numGroupMembers / self.tSettings[eOption_NumColumns])
	wndGroup:SetAnchorOffsets(nLeft, nTop, nLeft + nGroupWidth, nTop + nGroupHeight + 15)
	
	if numGroupMembers > 0 then		
		if bWidthCalc then
			local arMemberItems = wndGroup:FindChild("MemberList"):GetChildren()
			for i = 1, #arMemberItems do
				local nLeft, nTop, nRight, nBottom = arMemberItems[i]:GetAnchorOffsets()
				arMemberItems[i]:SetAnchorOffsets(0, nTop, self.tSettings[eOption_FrameWidth], nBottom)
				if self.tSettings[eOption_NumColumns] > 1 then
					arMemberItems[i]:FindChild("Contents"):SetAnchorOffsets(0, 0, -1 * knFramePadding, 1)
				else
					arMemberItems[i]:FindChild("Contents"):SetAnchorOffsets(0, 0, 0, 1)
				end
			end
		end
		wndGroup:FindChild("MemberList"):ArrangeChildrenTiles(0)
	end
	
	self:UpdateGroupTitle(wndGroup)
end

function QuickFrames:UpdateGroupTitle(wndGroup)
	local tGroupMembers = wndGroup:FindChild("MemberList"):GetChildren()
	local nNumGroupMembrs = #tGroupMembers
	local wndGroupName = wndGroup:FindChild("GroupName")
	local strTitle = wndGroupName:GetData()
	if nNumGroupMembrs > 0 then
		local nAlive = 0
		for k, wndMemberItem in pairs(tGroupMembers) do
			if self.tRaidMembers[wndMemberItem:FindChild("Contents"):GetData()].eStatus == ePlayerStatus_Alive then
				nAlive = nAlive + 1
			end
		end
		strTitle = strTitle .. " (" .. nAlive .. "/" .. nNumGroupMembrs .. ")"
	end
	wndGroupName:SetText(strTitle)
end

function QuickFrames:CreateGroup(strName)
	local strGroupName = strName
	
	local wndNew = Apollo.LoadForm(self.xmlDoc, "GroupForm", self.wndRaidFrame, self)
	wndNew:FindChild("GroupName"):SetData(strGroupName)	
	self:FormatGroup(wndNew, true)
	
	return wndNew
end

function QuickFrames:DestroyAll()
	for i = 1, #self.tGroups do
		self.tGroups[i]:Destroy()
	end
	self.tGroups = { }
	self.tRaidMembers = { }
	self.tSortedMembers = { }
	self.tCurrentlyMarked = { }
	self.updateTimer:Stop()
	self.readyCheckTimer:Stop()
	self.hideReadyCheckTimer:Stop()
	
	self.wndQuickActionsBtn:Show(false, true)
end

function QuickFrames:InitializeRaidGroup()
	self:DebugPrint("Starting Update Timer")
	self.updateTimer:Start()
	if Apollo.GetAddon("QuickFrames_Interrupts") then self.bShowInterrupts = true end
	self:BuildDefaultFrames()
	self.wndQuickActionsBtn:Show(self.tSettings[eOption_ShowQuickActions], true)
	self.wndRaidFrame:SetStyle("Moveable", self.tSettings[eOption_Locked])
end

function QuickFrames:ToggleLock()
	self.tSettings[eOption_Locked] = not self.tSettings[eOption_Locked]
	self.wndRaidFrame:SetStyle("Moveable", self.tSettings[eOption_Locked])

end

function QuickFrames:ShowLootOptions()
	Event_FireGenericEvent("GenericEvent_Raid_ToggleMasterLoot", true)
end

function QuickFrames:GiveRaidAssistToAll()
	for i = 1, GroupLib.GetMemberCount() do
		GroupLib.SetRaidAssistant(i, true)
	end
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Periodic
-----------------------------------------------------------------------------------------------
function QuickFrames:OnTimer()
	for i = 1, GroupLib.GetMemberCount() do
		curMember = GroupLib.GetGroupMember(i)
		local tCurMemberData = self.tRaidMembers[curMember.strCharacterName]
		local eStatus = ePlayerStatus_Alive
		if curMember.bDisconnected then
			eStatus = ePlayerStatus_Offline
		elseif curMember.nHealth == 0 and curMember.nHealthMax ~= 0 then
			eStatus = ePlayerStatus_Dead
		end
			
		if eStatus ~= tCurMemberData.eStatus then
			tCurMemberData.eStatus = eStatus
			if eStatus == ePlayerStatus_Offline then
				if tCurMemberData.bVitalsShown then self:HideVitals(tCurMemberData) end
				tCurMemberData.wndBGBar:SetBGColor(kColorOffline)
				tCurMemberData.wndNameText:SetTextColor(kTextColorOffline)
				tCurMemberData.wndMemberItem:FindChild("BuffStatusFood"):FindChild("Icon"):Show(false, true)
				tCurMemberData.wndMemberItem:FindChild("BuffStatusBoost"):FindChild("Icon"):Show(false, true)
			elseif eStatus == ePlayerStatus_Alive  and not tCurMemberData.bVitalsShown then
				if self.tSettings[eOption_ClassColors] then
					tCurMemberData.wndBGBar:SetBGColor(ktIdToClassColors[tCurMemberData.eClassId])
					tCurMemberData.wndNameText:SetTextColor(kTextColorAlive)
				else
					tCurMemberData.wndBGBar:SetBGColor(kColorAlive)
					tCurMemberData.wndNameText:SetTextColor(kTextColorAlive)
				end
			else
				if tCurMemberData.bVitalsShown then self:HideVitals(tCurMemberData) end
				tCurMemberData.wndBGBar:SetBGColor(kColorDead)
				tCurMemberData.wndNameText:SetTextColor(kTextColorDead)
				tCurMemberData.wndMemberItem:FindChild("BuffStatusFood"):FindChild("Icon"):Show(false, true)
				tCurMemberData.wndMemberItem:FindChild("BuffStatusBoost"):FindChild("Icon"):Show(false, true)
			end
		end

		if eStatus == ePlayerStatus_Alive and tCurMemberData.bVitalsShown then
			self:UpdateVitals(tCurMemberData, curMember)
		end
		
		if eStatus == ePlayerStatus_Alive and self.tSettings[eOption_ShowBuffStatus] then
			self:UpdateBuffStatus(tCurMemberData, GroupLib.GetUnitForGroupMember(i))
		

			if not GroupLib.GetUnitForGroupMember(i) and not tCurMemberData.wndMemberItem:ContainsMouse() then
				tCurMemberData.wndBGBar:SetBGOpacity(0.3)
				tCurMemberData.wndBorder:SetBGOpacity(0.3)
				if tCurMemberData.bVitalsShown then
					tCurMemberData.wndHealthBar:SetBGOpacity(0.3)
					tCurMemberData.wndShieldBar:SetBGOpacity(0.3)
					tCurMemberData.wndAbsorbBar:SetBGOpacity(0.3)
				end
			elseif not tCurMemberData.wndMemberItem:ContainsMouse() then
				tCurMemberData.wndBGBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
				tCurMemberData.wndBorder:SetBGOpacity(1)
				if tCurMemberData.bVitalsShown then
					tCurMemberData.wndHealthBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
					tCurMemberData.wndShieldBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
					tCurMemberData.wndAbsorbBar:SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
				end
			end
		end
	end
	
	for i = 1, #self.tGroups do
		self:UpdateGroupTitle(self.tGroups[i])
	end
end

function QuickFrames:UpdateBuffStatus(tCurMemberData, unit)
	if not tCurMemberData or not unit then
		return
	end
	
	tCurMemberData.wndMemberItem:FindChild("BuffStatusFood"):FindChild("Icon"):Show(self:UnitHasBuff(unit, kFoodBuffStr), true)
	
	local bBoosted = false
	for k, v in pairs(kBoostStrList) do
		if self:UnitHasBuff(unit, v) then
			bBoosted = true
			break
		end
	end
	tCurMemberData.wndMemberItem:FindChild("BuffStatusBoost"):FindChild("Icon"):Show(bBoosted, true)
end

function QuickFrames:UpdateVitals(tRaidMember, unitPlayer)
	-- Set Background to Black
	tRaidMember.wndNameText:SetTextColor(kTextColorAlive)
	tRaidMember.wndBGBar:SetBGColor(kColorOffline)
	local nHealth = unitPlayer.nHealth / unitPlayer.nHealthMax
	local nShield = unitPlayer.nShield / unitPlayer.nShieldMax
	local nAbsorb = 0

	if unitPlayer.nAbsorptionMax > 0 then
		nAbsorb = unitPlayer.nAbsorption / unitPlayer.nAbsorptionMax -- Since it doesn't clear when the buff drops off
	end

	-- Bars
	local wndHealthBar = tRaidMember.wndHealthBar
	local wndShieldBar = tRaidMember.wndShieldBar
	local wndAbsorbBar = tRaidMember.wndAbsorbBar

	if self.tSettings[eOption_ClassColors] then
		wndHealthBar:SetBGColor(ktIdToClassColors[tRaidMember.eClassId])
	else
		wndHealthBar:SetBGColor(kColorAlive)
	end

	local nLeft, nTop, nRight, nBottom = tRaidMember.wndBGBar:GetAnchorOffsets()
	local nHeight = tRaidMember.wndBGBar:GetHeight()
	local nWidth = tRaidMember.wndBGBar:GetWidth()

	-- Set Offsets
	wndHealthBar:SetAnchorOffsets( nLeft, nTop, (nLeft + nWidth) * nHealth, nBottom )
	wndShieldBar:SetAnchorOffsets( nLeft, (nHeight * .75) - nBottom, (nLeft + nWidth) * nShield, nBottom )

	if nAbsorb > 0 then 
		wndAbsorbBar:SetAnchorOffsets( nLeft, nTop, (nLeft + nWidth) * nAbsorb, (nHeight * .60) - nBottom ) 
		wndAbsorbBar:Show(true, true)
	else
		wndAbsorbBar:Show(false, true)
	end

	-- Show Bars
	wndHealthBar:Show(true, true)
	wndShieldBar:Show(unitPlayer.nShieldMax > 0 and nShield > 0, true)
end

function QuickFrames:HideVitals(tRaidMember)
	tRaidMember.wndHealthBar:Show(false, true)
	tRaidMember.wndShieldBar:Show(false, true)
	tRaidMember.wndAbsorbBar:Show(false, true)
end

-----------------------------------------------------------------------------------------------
-- QuickFrames UI Events
-----------------------------------------------------------------------------------------------
function QuickFrames:OnChatInvoke(strShortcut, args)
	local argTokens = self:Tokenize(args)
	
	if #argTokens == 0 then
		self:ToggleSettings()
	end

	if argTokens[1] == "rc" or argTokens[1] == "ready" or argTokens[1] == "readycheck" then
		self:StartReadyCheck(self:JoinTokens(argTokens, 2))
	elseif argTokens[1] == "l" or argTokens[1] == "lock" then
		self:ToggleLock()
	elseif argTokens[1] == "fr" or argTokens[1] == "forcereset" or argTokens[1] == "reset" then
		self:Reset()
	elseif argTokens[1] == "rn" or argTokens[1] == "rename" then
		if #argTokens < 3 then
			self:InfoPrint("Improper Usage! Use /rf rn [rename] <Old Character Name> <New Character Name>")
		elseif #argTokens == 3 then
			self:RenameCharacter(argTokens[2], argTokens[3])
		else
			self:RenameCharacter(argTokens[2] .. " " ..argTokens[3], argTokens[4])
		end
	end
end

function QuickFrames:OnMemberItemHover(wndHandler, wndControl, x, y)
	if not wndHandler then
		return
	end
	wndHandler:FindChild("BG"):SetBGOpacity(self.tSettings[eOption_HighlightedOpacity])
	wndHandler:FindChild("HealthBar"):SetBGOpacity(self.tSettings[eOption_HighlightedOpacity])
	wndHandler:FindChild("ShieldBar"):SetBGOpacity(self.tSettings[eOption_HighlightedOpacity])
	wndHandler:FindChild("AbsorbBar"):SetBGOpacity(self.tSettings[eOption_HighlightedOpacity])
end

function QuickFrames:OnMemberItemUnhover(wndHandler, wndControl, x, y)
	if not wndHandler then
		return
	end
	wndHandler:FindChild("BG"):SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	wndHandler:FindChild("HealthBar"):SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	wndHandler:FindChild("ShieldBar"):SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	wndHandler:FindChild("AbsorbBar"):SetBGOpacity(self.tSettings[eOption_DefaultOpacity])
	wndHandler:FindChild("BorderHighlight"):Show(false, true)
end

function QuickFrames:OnMemberItemPress(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation)
	if not wndHandler then
		return
	end
	
	wndHandler:FindChild("BorderHighlight"):Show(true, true)
end

function QuickFrames:OnMemberItemSelected(wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY)
	local strName = wndHandler:GetData()
	if not wndHandler or not strName then
		return
	end
	
	wndHandler:FindChild("BorderHighlight"):Show(false, true)
	
	local nGroupMemberCount = GroupLib.GetMemberCount()
	local tMemberInfo = nil
	local unit = nil
	for i = 1, nGroupMemberCount do
		local curMember = GroupLib.GetGroupMember(i)
		if curMember.strCharacterName == strName then
			tMemberInfo = curMember
			unit = GroupLib.GetUnitForGroupMember(i)
		end
	end
	
	if not tMemberInfo then
		self:DebugPrint("Group Member came back null for: " .. strName)
		return
	end
	
	self:DebugPrint("Selected: " .. tMemberInfo.strCharacterName)
	
	if eMouseButton == GameLib.CodeEnumInputMouse.Right then
		Event_FireGenericEvent("GenericEvent_NewContextMenuPlayerDetailed", wndHandler, tMemberInfo.strCharacterName, unit) -- unit is optional
	elseif unit then
		GameLib.SetTargetUnit(unit)
	end
end

function QuickFrames:OnDefaultRaidLeaderOptions()
	Event_FireGenericEvent("GenericEvent_Raid_ToggleLeaderOptions", true)
end

function QuickFrames:OnQuickActionsBtn( wndHandler, wndControl, eMouseButton )
	self:ShowQuickActionsContextMenu()
end

function QuickFrames:OnQuickActionsMenuBtn( wndHandler, wndControl, eMouseButton )
	self:HandleQuickAction(wndHandler:GetData())
	self.wndQuickActionsMenu:Destroy()
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Context Menu
-----------------------------------------------------------------------------------------------
function QuickFrames:ShowQuickActionsContextMenu()
	self.wndQuickActionsMenu = Apollo.LoadForm(self.xmlDoc, "ContextMenuForm", "TooltipStratum", self)
	local wndItemList = self.wndQuickActionsMenu:FindChild("ButtonList")
	
	for k, v in pairs(kQuickActionsMap) do
		if not v.bLeaderOnly or GroupLib.AmILeader() then
			if v.tSubMenuMap then
				local wndNew = self:AddContextMenuItem(wndItemList, k, v.strLabel, true)
				self:CreateSubMenu(wndNew, v.tSubMenuMap)
			else
				self:AddContextMenuItem(wndItemList, k, v.strLabel)
			end
		end
	end
	
	local tCursor = Apollo.GetMouse()
	self.wndQuickActionsMenu:Move(tCursor.x - knXCursorOffset, tCursor.y - knYCursorOffset, self.wndQuickActionsMenu:GetWidth(), self.wndQuickActionsMenu:GetHeight())
	local nHeight = wndItemList:ArrangeChildrenVert(0)
	local nLeft, nTop, nRight, nBottom = self.wndQuickActionsMenu:GetAnchorOffsets()
	self.wndQuickActionsMenu:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight + 62)
end

function QuickFrames:AddContextMenuItem(wndParent, eOptionId, strLabel, bMulti)
	wndNew = Apollo.LoadForm(self.xmlDoc, bMulti and "BtnMulti" or "BtnRegular", wndParent, self)
	wndNew:SetData(eOptionId)
	wndNew:FindChild("BtnText"):SetText(strLabel)
	return wndNew
end

function QuickFrames:CreateSubMenu(wndParent, tSubMenuMap)
	local wndSubMenu = wndParent:FindChild("PopoutFrame")
	local wndItemList = wndSubMenu:FindChild("PopoutItems")
	wndParent:AttachWindow(wndSubMenu)
	
	for k, v in pairs(tSubMenuMap) do
		if not v.bLeaderOnly or GroupLib.AmILeader() then
			self:AddContextMenuItem(wndItemList, k, v.strLabel)
		end
	end
	
	local nHeight = wndItemList:ArrangeChildrenVert(0)
	local nLeft, nTop, nRight, nBottom = wndSubMenu:GetAnchorOffsets()
	wndSubMenu:SetAnchorOffsets(nLeft, nTop, nRight, nTop + nHeight + 62)
	
	return wndSubMenu
end

function QuickFrames:HandleQuickAction(eOptionId)
	if eOptionId == eQuickActionsMenu_ReadyCheck then
		self:StartReadyCheck()
	elseif eOptionId == eQuickActionsMenu_LootOptions then
		self:ShowLootOptions()
	elseif eOptionId == eQuickActionsMenu_LeaderOptions then
		self:OnDefaultRaidLeaderOptions()
	elseif eOptionId == eQuickActionsMenu_Lock then
		self:ToggleLock()
	elseif eOptionId == eQuickActionsMenu_MassRaidAssist then
		self:GiveRaidAssistToAll()
	elseif eOptionId == eQuickActionsMenu_RoleDPS then
		GroupLib.SetRoleDPS(self:GroupIndexByName(GameLib.GetPlayerUnit():GetName()), true)
	elseif eOptionId == eQuickActionsMenu_RoleHealer then
		GroupLib.SetRoleHealer(self:GroupIndexByName(GameLib.GetPlayerUnit():GetName()), true)
	elseif eOptionId == eQuickActionsMenu_RoleTank then
		GroupLib.SetRoleTank(self:GroupIndexByName(GameLib.GetPlayerUnit():GetName()), true)
	end
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Game Events
-----------------------------------------------------------------------------------------------
function QuickFrames:OnGroupFlagsChanged()
	if GroupLib.InRaid() then
		self:InitializeRaidGroup()
		
		local bLeader = GroupLib.AmILeader()
		local wndOptions = self.wndSettings:FindChild("OptionsFrame")
		wndOptions:FindChild("RaidLeaderOptions"):Enable(bLeader)
		wndOptions:FindChild("LootOptions"):Enable(bLeader)
	end
end

function QuickFrames:OnGroupJoin()
	if GroupLib.InRaid() then
		self:InitializeRaidGroup()
	end
end

function QuickFrames:OnGroupLeft()
	self.nNumReadyCheckResponses = -1
	self:DestroyAll()
end

function QuickFrames:OnGroupAdd(strName)
	if self.tRaidMembers[strName] ~= nil then
		return
	end
	
	for i = 1, GroupLib.GetMemberCount() do
		local tCurMember = GroupLib.GetGroupMember(i)
		if tCurMember.strCharacterName == strName then
			local tMemberData = self:CreateMemberData(tCurMember)
			self.tRaidMembers[strName] = tMemberData
			break
		end
	end
	self:UpdateMemberSort()
end

function QuickFrames:OnGroupRemove(strName)
	local lowerName = strName
	if self.tRaidMembers[strName] then
		self:RemoveMemberFromGroup(self.tRaidMembers[strName])
		self.tRaidMembers[strName] = nil
	end

	for key, name in pairs(self.tSortedMembers) do
		if name[1] == lowerName then
			table.remove(self.tSortedMembers, key)
		end
	end		
	
	for key, name in pairs(self.tCurrentlyMarked) do
		if name == lowerName then
			self.tCurrentlyMarked[key] = nil
			break
		end
	end
	self:UpdateMemberSort()
end

function QuickFrames:OnSetMark(nMarkIdx, unit)
	if self.tCurrentlyMarked[nMarkIdx] then
		self.tRaidMembers[self.tCurrentlyMarked[nMarkIdx]].wndMemberItem:FindChild("Mark"):FindChild("Icon"):Show(false, true)
		self.tCurrentlyMarked[nMarkIdx] = nil
	end
	
	if unit then
		local lowerName = unit:GetName()
		if not self.tRaidMembers[lowerName] then
			return
		end
		
		self.tCurrentlyMarked[nMarkIdx] = lowerName
		local wndMark = self.tRaidMembers[lowerName].wndMemberItem:FindChild("Mark"):FindChild("Icon")
		wndMark:SetSprite(kstrRaidMarkerToSprite[nMarkIdx])
		wndMark:Show(true, true)
	end
end

function QuickFrames:OnMemberPromoted(strNewLeader)
	if not GroupLib.InRaid() then return end
	self:DestroyAll()
	self:InitializeRaidGroup()
	local bLeader = GroupLib.AmILeader()
	local wndOptions = self.wndSettings:FindChild("OptionsFrame")
	wndOptions:FindChild("RaidLeaderOptions"):Enable(bLeader)
	wndOptions:FindChild("LootOptions"):Enable(bLeader)
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Ready Check
-----------------------------------------------------------------------------------------------
function QuickFrames:StartReadyCheck(strMessage)
	if self.nNumReadyCheckResponses ~= -1 then
		return
	end
	
	if strMessage == nil then
		strMessage = Apollo.GetString("RaidFrame_AreYouReady")
	end
	GroupLib.ReadyCheck(strMessage)
end

function QuickFrames:OnReadyCheck(nMemberIdx, strMessage)
	self:HideReadyCheckIcons()
	
	local tMember = GroupLib.GetGroupMember(nMemberIdx)
	local strName = Apollo.GetString("RaidFrame_TheRaid")
	if tMember then
		strName = tMember.strCharacterName
	end

	if self.wndReadyCheckPopup and self.wndReadyCheckPopup:IsValid() then
		self.wndReadyCheckPopup:Destroy()
	end

	self.wndReadyCheckPopup = Apollo.LoadForm(self.xmlDoc, "RaidReadyCheck", nil, self)
	self.wndReadyCheckPopup:FindChild("ReadyCheckYesBtn"):SetData(1)
	self.wndReadyCheckPopup:FindChild("ReadyCheckMessage"):SetText(String_GetWeaselString(Apollo.GetString("RaidFrame_ReadyCheckStarted"), strName) .. "\n" .. strMessage)

	self.nNumReadyCheckResponses = 0

	self.readyCheckTimer:Start()
end

function QuickFrames:OnReadyCheckResponse(wndHandler, wndControl)
	if wndHandler == wndControl then
		GroupLib.SetReady(wndHandler:GetData() ~= nil)
	end

	if self.wndReadyCheckPopup and self.wndReadyCheckPopup:IsValid() then
		self.wndReadyCheckPopup:Destroy()
	end
end

function QuickFrames:OnMemberFlagsChanged(nMemberIdx, bFromPromotion, tChangedFlags)
	if not GroupLib.InRaid() then return end

	local tMember = GroupLib.GetGroupMember(nMemberIdx)
	local bReady = tChangedFlags.bReady
	local tMemberData = self.tRaidMembers[tMember.strCharacterName]

	if tChangedFlags.bRaidAssistant then
		tMemberData.bIsRaidAssist = tMember.bRaidAssistant
		if tMemberData.bIsRaidAssist and not tMemberData.bIsLeader then 
			tMemberData.wndBorder:SetSprite("CRB_ActionBarIconSprites:sprActionBar_OrangeBorder") 
			tMemberData.wndBorder:SetBGColor(CColor.new(.2,.2,0))
		elseif not tMemberData.bIsLeader then
			tMemberData.wndBorder:SetSprite("CRB_Basekit:kitBase_HoloBlue_InsetBorder_Thin") 
			tMemberData.wndBorder:SetBGColor(CColor.new(0,0,0))
		end
	end

	if tChangedFlags.bTank or tChangedFlags.bHealer or tChangedFlags.bDPS then
		tMemberData.nRole = self:GetRole(tMember)
		self:UpdateMemberSort()
		return
	end

	if self.nNumReadyCheckResponses == -1 then
		return
	end
	
	if tMemberData then
		tMemberData.bReady = bReady
		
		local wndReadyIcon = tMemberData.wndMemberItem:FindChild("ReadyIcon")
		wndReadyIcon:SetSprite(bReady and kReadySprite or kNotReadySprite)
		wndReadyIcon:Show(true, true)
		self:AdjustBuffPositions(tMemberData.wndMemberItem, true)
	end
	
	self.nNumReadyCheckResponses = self.nNumReadyCheckResponses + 1
	if self.nNumReadyCheckResponses == GroupLib.GetMemberCount() then
		self:OnReadyCheckTimeout()
	end
end

function QuickFrames:OnReadyCheckTimeout()
	self.readyCheckTimer:Stop()
	
	self.nNumReadyCheckResponses = -1

	if self.wndReadyCheckPopup and self.wndReadyCheckPopup:IsValid() then
		self.wndReadyCheckPopup:Destroy()
	end

	local strMembersNotReady = ""
	for k, tMemberData in pairs(self.tRaidMembers) do
		if not tMemberData.bReady then
			local wndReadyIcon = tMemberData.wndMemberItem:FindChild("ReadyIcon")
			wndReadyIcon:SetSprite(kNotReadySprite)
			wndReadyIcon:Show(true, true)
			self:AdjustBuffPositions(tMemberData.wndMemberItem, true)
			
			if strMembersNotReady == "" then
				strMembersNotReady = tMemberData.strCharacterName
			else
				strMembersNotReady = String_GetWeaselString(Apollo.GetString("RaidFrame_NotReadyList"), strMembersNotReady, tMemberData.strCharacterName)
			end
		end
		
		tMemberData.bReady = false
	end

	if strMembersNotReady == "" then
		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Party, Apollo.GetString("RaidFrame_ReadyCheckSuccess"), "")
	else
		ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_Party, String_GetWeaselString(Apollo.GetString("RaidFrame_ReadyCheckFail"), strMembersNotReady), "")
	end
	
	self.hideReadyCheckTimer:Start()
end

function QuickFrames:HideReadyCheckIcons()
	for k, tMemberData in pairs(self.tRaidMembers) do
		tMemberData.wndMemberItem:FindChild("ReadyIcon"):Show(false, true)
		self:AdjustBuffPositions(tMemberData.wndMemberItem, false)
	end
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Helpers
-----------------------------------------------------------------------------------------------
function QuickFrames:Tokenize(str)
	local tokens = {}
	for token in string.gmatch(str, "%S+") do
		table.insert(tokens, token)
	end
	return tokens
end

function QuickFrames:JoinTokens(arTokens, nStartIdx)
	if nStartIdx > #arTokens then
		return nil
	end
	
	local strJoined = arTokens[nStartIdx]
	for i = nStartIdx + 1, #arTokens do
		strJoined = strJoined .. " " .. arTokens[i]
	end
	return strJoined
end

function QuickFrames:CreateMemberData(newMember)
	local tData = {
		eStatus = ePlayerStatus_Alive,
		strCharacterName = newMember.strCharacterName,
		eClassId = newMember.eClassId,
		nMarkerId = newMember.nMarkerId,
		bReady = false,
		bVitalsShown = self.tSettings[eOption_ShowVitals],
		nRole = self:GetRole(newMember),
		bInRange = true,
		bIsLeader = newMember.bIsLeader,
		bIsRaidAssist = newMember.bRaidAssistant
	}

	table.insert(self.tSortedMembers, {newMember.strCharacterName, newMember.eClassId} )
	
	if newMember.bDisconnected then
		tData.eStatus = ePlayerStatus_Offline
	elseif newMember.nHealth == 0 and newMember.nHealthMax ~= 0 then
		tData.eStatus = ePlayerStatus_Dead
	end
	
	return tData
end

function QuickFrames:GetRole(tMember)
	if tMember.bTank then
		return 1
	elseif tMember.bHealer then
		return 2
	else
		return 3
	end
end

function QuickFrames:DebugPrint(str)
	if kDebug and str then
		Print("[QF] " .. str)
	end
end

function QuickFrames:InfoPrint(str)
	ChatSystemLib.PostOnChannel(ChatSystemLib.ChatChannel_System, str, "QuickFrames")
end

function QuickFrames:UnitHasBuff(unit, strBuff)
	local tBuffs = unit:GetBuffs().arBeneficial
	if not tBuffs then
		return false
	end
	
	for k, v in pairs(tBuffs) do
		if v.splEffect:GetName() == strBuff then
			return true
		end
	end
	return false
end

function QuickFrames:AdjustBuffPositions(wndMember, bReadyCheckIconShown)
	if not wndMember then
		return
	end
	
	local wndFoodBuff = wndMember:FindChild("BuffStatusFood")
	local wndBoost = wndMember:FindChild("BuffStatusBoost")
	
	local nShift = bReadyCheckIconShown and knReadyCheckWidth or 0
	
	nLeft, nTop, nRight, nBottom = wndFoodBuff:GetAnchorOffsets()
	wndFoodBuff:SetAnchorOffsets(knFoodBuffOffset - wndFoodBuff:GetWidth() - nShift, nTop, knFoodBuffOffset - nShift, nBottom)
	
	nLeft, nTop, nRight, nBottom = wndBoost:GetAnchorOffsets()
	wndBoost:SetAnchorOffsets(knBoostOffset - wndBoost:GetWidth() - nShift, nTop, knBoostOffset - nShift, nBottom)
end

function QuickFrames:GroupIndexByName(strName)
	for i = 1, GroupLib.GetMemberCount() do
		if strName == GroupLib.GetGroupMember(i).strCharacterName then
			return i
		end
	end
	return 0
end

---------------------------------------------------------------------------------------------------
-- QuickFrames Settings
---------------------------------------------------------------------------------------------------
function QuickFrames:ToggleSettings()
	self.wndSettings:Show(not self.wndSettings:IsShown())
end

function QuickFrames:OnCloseSettings(wndHandler, wndControl, eMouseButton)
	self.wndSettings:Show(false)
end

function QuickFrames:OnOptionCheckbox(wndHandler, wndControl, eMouseButton)
	local eOption = wndHandler:GetData()
	local bValue = wndHandler:IsChecked()
	self.tSettings[eOption] = bValue
	self:OptionToggled(eOption, bValue)
end

function QuickFrames:OptionToggled(eOption, bValue)
	if eOption == eOption_ShowMarks then
		for k, tMemberData in pairs(self.tRaidMembers) do
			tMemberData.wndMemberItem:FindChild("Mark"):Show(bValue, true)
		end
	elseif eOption == eOption_ShowBuffStatus then
		for k, tMemberData in pairs(self.tRaidMembers) do
			tMemberData.wndMemberItem:FindChild("BuffStatusFood"):Show(bValue, true)
			tMemberData.wndMemberItem:FindChild("BuffStatusBoost"):Show(bValue, true)
		end
	elseif eOption == eOption_ShowQuickActions then
		self.wndQuickActionsBtn:Show(bValue, true)
	elseif eOption == eOption_ShowVitals then
		for k, tMemberData in pairs(self.tRaidMembers) do
			tMemberData.bVitalsShown = bValue
			if not tMemberData.bVitalsShown then 
				self:HideVitals(tMemberData) 
				tMemberData.eStatus = -1 -- To force BG to update
			end
		end
		self:OnTimer()
	elseif eOption == eOption_ClassColors then
		for k, tMemberData in pairs(self.tRaidMembers) do
			if bValue and tMemberData.eStatus == ePlayerStatus_Alive then
				tMemberData.wndMemberItem:FindChild("BG"):SetBGColor(ktIdToClassColors[tMemberData.eClassId])
			elseif tMemberData.eStatus == ePlayerStatus_Alive then
				tMemberData.wndMemberItem:FindChild("BG"):SetBGColor(kColorAlive)
			end
		end 
	end
end

function QuickFrames:OnOptionsSliderChanged(wndHandler, wndControl, fValue, fOldValue)
	local nDecimalPlaces = kSettingsDataMap[wndHandler:GetParent():GetName()].bInteger and 0 or 2
	wndHandler:GetParent():FindChild("EditBox"):SetText(string.format("%." .. nDecimalPlaces .. "f", fValue))
	local eOption = wndHandler:GetParent():GetData()
	self.tSettings[wndHandler:GetParent():GetData()] = fValue
	
	if eOption == eOption_DefaultOpacity then
		for k, tMemberData in pairs(self.tRaidMembers) do
			tMemberData.wndMemberItem:FindChild("BG"):SetBGOpacity(fValue)
			tMemberData.wndMemberItem:FindChild("HealthBar"):SetBGOpacity(fValue)
			tMemberData.wndMemberItem:FindChild("ShieldBar"):SetBGOpacity(fValue)
			tMemberData.wndMemberItem:FindChild("AbsorbBar"):SetBGOpacity(fValue)
		end
	elseif eOption == eOption_FrameWidth or eOption == eOption_NumColumns then
		self:FormatGroups()
	elseif eOption == eOption_UpdateTimer then
		self.updateTimer:Set(self.tSettings[eOption_UpdateTimer], true, "OnTimer")
	end
end

-----------------------------------------------------------------------------------------------
-- QuickFrames Instance
-----------------------------------------------------------------------------------------------
local QuickFramesInst = QuickFrames:new()
QuickFramesInst:Init()
